#include "repetition_testing/reptester.h"
#include "profiler/timer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void handle_alloc(reptester *tester, alloc_type type) {
  switch (type) {
  case ALLOC_TYPE_WITH_MALLOC:
    if (!(tester->params.buffer)) {
      tester->params.buffer = (char *)malloc(tester->params.read_size + 1);
      memset(tester->params.buffer, 0, tester->params.read_size + 1);
    }

    break;
  default:
    break;
  }
}

void handle_free(reptester *tester, alloc_type type) {
  switch (type) {
  case ALLOC_TYPE_WITH_MALLOC:
    if (tester->params.buffer) {
      free(tester->params.buffer);
      tester->params.buffer = NULL;
    }

    break;
  default:
    break;
  }
}

void run_func_test(reptester *tester, reptest_func func, const char *func_name,
                   alloc_type type) {
  tester->test_start_time = read_cpu_timer();
  tester->test_time_secs = 0.0;
  tester->current_run = 1;
  tester->min = UINT64_MAX;
  tester->max = 0;
  tester->avg = 0;
  tester->total = 0;
  tester->results = {};

  if (type == ALLOC_TYPE_WITHOUT_MALLOC && !(tester->params.buffer)) {
    tester->params.buffer = (char *)malloc(tester->params.read_size + 1);
    memset(tester->params.buffer, 0, tester->params.read_size + 1);
  }

  while (tester->test_time_secs <= tester->wait_time_secs) {
    func(tester, type);

    if (tester->results.bytes_read <
        tester->params.read_size * tester->params.read_count) {
      printf("Failed to read the entire file (Total size: %lu, Bytes read: "
             "%lu)\n",
             tester->params.read_size, tester->results.bytes_read);

      return;
    }

    tester->total += tester->results.read_time;

    if (tester->results.read_time > tester->max) {
      tester->max = tester->results.read_time;
    } else if (tester->results.read_time < tester->min) {
      tester->test_start_time = read_cpu_timer();
      tester->min = tester->results.read_time;
    }

    tester->test_time_secs = time_in_seconds(
        read_cpu_timer() - tester->test_start_time, tester->cpu_freq);

    ++(tester->current_run);
  }

  if (type == ALLOC_TYPE_WITHOUT_MALLOC && tester->params.buffer) {
    free(tester->params.buffer);
    tester->params.buffer = NULL;
  }

  print_results(tester, func_name);
}

void print_results(reptester *tester, const char *name) {
  f64 gb = 1024.0 * 1024.0 * 1024.0;

  f64 size_in_gb =
      (f64)(tester->params.read_size * tester->params.read_count) / gb;

  u64 run_count = tester->current_run - 1;

  tester->avg = tester->total / run_count;

  printf("\n%s: %lu runs\n", name, run_count);
  printf("MIN: %lu (%fGB/s)\n", tester->min,
         size_in_gb / time_in_seconds(tester->min, tester->cpu_freq));
  printf("MAX: %lu (%fGB/s)\n", tester->max,
         size_in_gb / time_in_seconds(tester->max, tester->cpu_freq));
  printf("AVG: %lu (%fGB/s)\n", tester->avg,
         size_in_gb / time_in_seconds(tester->avg, tester->cpu_freq));
}
